/*-------------------------------------------------------------------------+
| Copyright 2020 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view.fx;

import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.toList;
import static org.fortiss.tooling.base.ui.annotation.view.fx.FXAnnotationFilter.HIERARCHY_LEVELS_ALL;
import static org.fortiss.tooling.base.ui.annotation.view.fx.FXAnnotationFilter.HIERARCHY_LEVELS_CURRENT;
import static org.fortiss.tooling.base.ui.annotation.view.fx.FXAnnotationFilter.HIERARCHY_LEVELS_SELECTED_SUBMODEL;
import static org.fortiss.tooling.kernel.ui.util.SelectionUtils.checkAndPickFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstParentWithType;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.showWarning;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.annotation.AnnotationValueService;
import org.fortiss.tooling.base.annotation.IAnnotationValueService;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureDescriptor;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.annotation.valueprovider.ValueProviderBase;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.view.generic.ColumnHandle;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTextFieldTreeTableCell;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeContentProviderBase;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeTableUIProviderBase;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeTableViewer;
import org.fortiss.tooling.common.ui.javafx.layout.CompositeFXControllerBase;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.CheckBox;
import javafx.scene.control.ComboBox;
import javafx.scene.control.RadioButton;
import javafx.scene.control.SplitPane;
import javafx.scene.control.TextField;
import javafx.scene.control.TreeItem;
import javafx.scene.control.TreeTableColumn;
import javafx.scene.control.TreeTableView;

/**
 * 
 * @author bayha
 */
@SuppressWarnings("unchecked")
public class AnnotationViewFXController extends CompositeFXControllerBase<SplitPane, Node>
		implements ISelectionListener {

	@FXML
	TextField txtFilterText;

	@FXML
	RadioButton radBtnFilterModelElements;

	@FXML
	RadioButton radBtnFilterAnnotationNames;

	@FXML
	CheckBox chkBoxMatchCase;

	@FXML
	CheckBox chkBoxselectedTypeOnly;

	@FXML
	ComboBox<String> comboHierarchy;

	@FXML
	ComboBox<String> comboAnnotationType;

	/**  */
	@FXML
	private TreeTableView<AnnotationEntry> annotationTreeTableView;

	/**  */
	private DynamicTreeTableViewer<AnnotationEntry> annotationViewer;

	/**  */
	private IProjectRootElement curentRootElement;
	
	IModelElement selected;
	
	private FXAnnotationFilter filter = new FXAnnotationFilter();

	private Set<ColumnHandle<IAnnotatedSpecification>> columnHandles = new TreeSet<>();

	private Map<ColumnHandle<IAnnotatedSpecification>, TreeTableColumn<AnnotationEntry, ?>> columnHandleToColumn = new HashMap<ColumnHandle<IAnnotatedSpecification>, TreeTableColumn<AnnotationEntry, ?>>();

	private Map<String, IAnnotatedSpecification> colNameToSpecification = new HashMap<String, IAnnotatedSpecification>();

	/**
	 * Associates the handles of the annotations with their ordered column indexes.
	 * Used for the UI providers of cell items.
	 */
	Map<Integer, ColumnHandle<IAnnotatedSpecification>> colIdxAnnotationMap = new HashMap<>();;

	/** {@inheritDoc} */
	@Override
	public String getFXMLLocation() {
		return "AnnotationViewFx.fxml";
	}

	/** {@inheritDoc} */
	@Override
	public void initialize() {
		setupFilterWidgets();
	}

	/** {@inheritDoc} */
	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		EObject selected = checkAndPickFirst(selection, EObject.class);

		if (selected == null) {
			clearTableViewer();
			return;
		}

		if (selected instanceof IProjectRootElement) {
			curentRootElement = (IProjectRootElement) selected;
		} else {
			curentRootElement = getFirstParentWithType(selected, IProjectRootElement.class);
		}

		if (curentRootElement == null) {
			// There was no root element for the selected element. (Usually should be the
			// FileProject.) This is not displayed here.
			clearTableViewer();

			return;
		}

		if (selected instanceof IModelElement) {
			filter.setCurrentySelectedModelElement((IModelElement) selected);
			this.selected = (IModelElement) selected;
		}

		ObservableList<String> annotationTypeOptions = comboAnnotationType.getItems();
		annotationTypeOptions.add("Show all annotation types.");
		annotationTypeOptions.addAll(columnHandles.stream().map(h -> h.getColumnName()).collect(toList()));
		
		setupAnnotationTree();
	}

	/**
	 * Generates / restructures the annotation tree according to the currently
	 * selectedElement.
	 */
	private void setupAnnotationTree() {
		if (curentRootElement == null) {
			return;
		}

		clearTableViewer();

		// Collect the annotation entries for all elements.
		EList<IModelElement> childElements = getChildrenWithType(curentRootElement, IModelElement.class);

		ArtificialRoot rootElement = new ArtificialRoot();

		for (IModelElement elem : childElements) {
			AnnotationEntry entry = IAnnotationValueService.getInstance().getAnnotationEntry(elem);

			if (entry != null && !entry.getSpecificationsList().isEmpty()) {
				rootElement.elements.add(entry);
			}
		}

		Set<AnnotationEntry> annotationEntries = new HashSet<>();
		childElements.forEach(e -> annotationEntries.add(AnnotationValueService.getInstance().getAnnotationEntry(e)));

		// Aggregate required columns. Column order is defined by
		// ColumnHandle.compareTo().
		for (AnnotationEntry entry : annotationEntries) {
			for (IAnnotatedSpecification spec : entry.getSpecificationsList()) {
				ColumnHandle<IAnnotatedSpecification> columnHandle = new ColumnHandle<IAnnotatedSpecification>(entry,
						spec);
				columnHandles.add(columnHandle);
				colNameToSpecification.put(columnHandle.getColumnName(), spec);
			}
		}

		// The DynamicTreeTableViewer needs to be created new, as the root is different.
		boolean showRoot = false;
		int revealLevel = 1;
		DynamicTreeTableUIProviderBase<AnnotationEntry> uiProvider = new AnnotationTreeTableUIProvider();
		DynamicTreeContentProviderBase<AnnotationEntry> contentProvider = new AnnotationTableContentProvider();
		annotationViewer = new DynamicTreeTableViewer<AnnotationEntry>(annotationTreeTableView, rootElement, showRoot,
				revealLevel, contentProvider, uiProvider);

		annotationViewer.addColumn("Model Element", 200);
		annotationViewer.addColumn("Comment", 250);

		int columnIdx = 2;
		for (ColumnHandle<IAnnotatedSpecification> handle : columnHandles) {
			TreeTableColumn<AnnotationEntry, ?> col = createColumnForHandle(columnIdx, handle);
			columnHandleToColumn.put(handle, col);

			columnIdx++;
		}
		filterColumns();
	}

	private void clearTableViewer() {
//		colIdxAnnotationMap.clear();
//		columnHandles.clear();
//		columnHandleToColumn.clear();
//		annotationTreeTableView.getColumns().clear();
//		annotationTreeTableView.setRoot(null);
	}

	private void setupFilterWidgets() {
		ObservableList<String> hierarchyLevels = comboHierarchy.getItems();
		hierarchyLevels.clear();
		hierarchyLevels.add(HIERARCHY_LEVELS_ALL);
		hierarchyLevels.add(HIERARCHY_LEVELS_CURRENT);
		hierarchyLevels.add(HIERARCHY_LEVELS_SELECTED_SUBMODEL);
		comboHierarchy.setValue(HIERARCHY_LEVELS_ALL);

		comboHierarchy.valueProperty().addListener((obs, oVal, nVal) -> {
			filter.setHierarchyLevelFilter(nVal);

			filterColumns();
			updateAnnotationTree();
		});

		comboAnnotationType.valueProperty().addListener((obs, oVal, nVal) -> {
			filter.setHierarchyLevelFilter(nVal);

			updateAnnotationTree();
		});

		txtFilterText.textProperty().addListener((obs, oVal, nVal) -> {
			filter.setNameFilterPattern(nVal);

			filterColumns();

			updateAnnotationTree();
		});

		radBtnFilterAnnotationNames.selectedProperty().addListener((obs, oVal, nVal) -> {
			filter.setFilterColumnName(nVal);

			updateAnnotationTree();
		});

		chkBoxMatchCase.selectedProperty().addListener((obs, oVal, nVal) -> {
			filter.setFilterNameMatchCase(nVal);

			updateAnnotationTree();
		});

		chkBoxselectedTypeOnly.selectedProperty().addListener((obs, oVal, nVal) -> {
			filter.setRestrictToSelectedModelElementType(nVal);

			updateAnnotationTree();
		});

		comboAnnotationType.valueProperty().addListener((obs, oVal, nVal) -> {
			IAnnotatedSpecification selectedSpec = colNameToSpecification.get(nVal);
			if (selectedSpec != null) {
				filter.setAnnotationTypeFilter(selectedSpec.getClass());
			} else {
				// In this case, the filter has been reset to "All Types"
				filter.setAnnotationTypeFilter(null);
			}

			filterColumns();
			updateAnnotationTree();
		});
	}

	private void updateAnnotationTree() {
		annotationTreeTableView.refresh();
		
		annotationViewer.update();
		annotationViewer.expandAllItems();
	}

	private void filterColumns() {
		for (ColumnHandle<IAnnotatedSpecification> c : columnHandles) {
			TreeTableColumn<AnnotationEntry, ?> col = columnHandleToColumn.get(c);
			col.setVisible(filter.passesColumnFilter(c));
		}

		updateAnnotationTree();
	}

	/** Creates the annotation column for the given handle. */
	private TreeTableColumn<AnnotationEntry, ?> createColumnForHandle(int columnIdx,
			ColumnHandle<IAnnotatedSpecification> handle) {
		Class<IAnnotatedSpecification> annotationClass = (Class<IAnnotatedSpecification>) handle
				.getAnnotatedSpecification().getClass();
		AnnotationEntry entry = handle.getEntry();
		IAnnotationValueProvider<IAnnotatedSpecification> valueProvider = entry
				.getAnnotationValueProvider(annotationClass);
		IAnnotatedSpecification specification = entry.getSpecification(annotationClass);
		EClassifier valueType = valueProvider.getEStructuralFeatureDescriptor().getEType(specification);

		String columnName = valueProvider.getAnnotationName(specification);

		colIdxAnnotationMap.put(columnIdx, handle);

		TreeTableColumn<AnnotationEntry, String> column = null;

		// Case distinction for the cell factory depending on the type to be edited.
		if (valueType == null) {
			column = annotationViewer.addColumn(columnName, 150);
			column.setEditable(false);

			return column;
		} else if (valueType instanceof EEnum) {
			// Add a combo column for enumerations.
			column = annotationViewer.addComboColumn(columnName, 150, rowEntry -> {
				ValueProviderBase<IAnnotatedSpecification> annotationValueProvider = (ValueProviderBase<IAnnotatedSpecification>) rowEntry
						.getAnnotationValueProvider(annotationClass);
				EStructuralFeatureDescriptor structuralFeature = annotationValueProvider
						.getEStructuralFeatureDescriptor();
				EClassifier type = structuralFeature.getEType(rowEntry.getSpecification(annotationClass));

				// Get the enum literals again, as it might be different for each row.
				if (type instanceof EEnum) {
					EList<EEnumLiteral> allLiterals = ((EEnum) type).getELiterals();

					return allLiterals.stream().map(l -> l.getLiteral()).collect(toList());
				}
				return emptyList();
			});
		} else if (valueType.getInstanceClass().equals(Boolean.class)) {
			// Use a checkbox column for booleans
			annotationViewer.addCheckboxColumn(columnName, 100);
		} else {
			// For all other types, there is text editing.
			column = annotationViewer.addColumn(columnName, 150);
			column.setCellFactory(ttColumn -> new DynamicTextFieldTreeTableCell<AnnotationEntry>() {
				protected boolean canEdit(AnnotationEntry element) {
					// We start editing only, if this cell is supposed to be editable.
					IAnnotationValueProvider<IAnnotatedSpecification> annotationValueProvider = element
							.getAnnotationValueProvider(annotationClass);
					IAnnotatedSpecification currentSpecification = element.getSpecification(annotationClass);
					return annotationValueProvider != null && annotationValueProvider.canEdit(currentSpecification);
				}

				protected void setValueOnLeavingCell(AnnotationEntry element, String currentValue) {
					IAnnotatedSpecification specification = element.getSpecification(annotationClass);
					setAnnotationEntryValue(specification, element, currentValue);
				}
			});
		}

		if (column != null) {
			column.setEditable(true);

			column.setOnEditCommit(event -> {
				TreeItem<AnnotationEntry> treeItem = event.getRowValue();
				AnnotationEntry rowEntry = treeItem.getValue();
				String newValue = event.getNewValue();

				setAnnotationEntryValue(specification, rowEntry, newValue);
			});
		}

		return column;
	}

	/**
	 * Local helper to set the given newValue for the given annotationEntry and
	 * specification.
	 */
	private void setAnnotationEntryValue(IAnnotatedSpecification specification, AnnotationEntry annotationEntry,
			String newValue) {
		ITopLevelElement tle = IPersistencyService.getInstance().getTopLevelElementFor(curentRootElement);

		tle.runAsCommand(() -> {
			try {
				annotationEntry.setSpecificationValue(newValue, specification.getClass());
			} catch (Exception e1) {
				showWarning("Problem during setting an AnnotationValue: \n" + e1.getStackTrace());
			}
		});
	}

	private String getAnnotationEntryName(AnnotationEntry entry) {
		IModelElement modelElement = entry.getModelElement();
		if (modelElement instanceof INamedCommentedElement) {
			return ((INamedCommentedElement) modelElement).getName();
		}
		return "";
	}

	/**
	 * UIProvider for the Annotation TreeTable.
	 */
	protected class AnnotationTreeTableUIProvider extends DynamicTreeTableUIProviderBase<AnnotationEntry> {

		/** {@inheritDoc} */
		@Override
		public String getLabel(AnnotationEntry entry, int column) {
			IModelElement element = entry.getModelElement();

			switch (column) {
			case 0:
				return getAnnotationEntryName(entry);
			case 1:
				if (element instanceof INamedCommentedElement) {
					return ((INamedCommentedElement) element).getComment();
				}
				return "";
			default:
				IAnnotatedSpecification columnSpec = colIdxAnnotationMap.get(column).getAnnotatedSpecification();

				Object specificationValue = entry.getSpecificationValue(columnSpec.getClass());
				if (specificationValue != null) {
					return specificationValue.toString();
				}

				return "";
			}
		}
	}

	/**
	 * Content provider, that can deal with the {@link ArtificialRoot}.
	 * 
	 * The tree under this root has only one layer.
	 */
	protected class AnnotationTableContentProvider extends DynamicTreeContentProviderBase<AnnotationEntry> {
		/** {@inheritDoc} */
		@Override
		protected Collection<? extends AnnotationEntry> getChildren(AnnotationEntry parent) {
			if (parent instanceof ArtificialRoot) {
				return ((ArtificialRoot) parent).elements;
			}

			return emptyList();
		}

		@Override
		protected boolean filter(AnnotationEntry entry, String filterValue) {
			return filter.passesRowFilter(entry);
		}
	}

	/**
	 * Local class to represent the invisible root for the tree table.
	 * 
	 * The only thing relevant here, is the {@link List} 'elements'.
	 */
	private class ArtificialRoot extends AnnotationEntry {

		/**
		 * Constructor.
		 */
		public ArtificialRoot() {
			super(selected);
		}

		/** The list of {@link AnnotationEntry}s to be displayed. */
		List<AnnotationEntry> elements = new ArrayList<AnnotationEntry>();
	}
}
